#!/usr/bin/python
#
# get changelogs, NEWS.Debian and copyright directly from LP
# instead of crawling the archive.
#

import apt_pkg
import datetime
import glob
import logging
import os
import urllib
import urllib2
import signal
import shutil
import subprocess
import sys
import tempfile
import time

from launchpadlib.launchpad import Launchpad, EDGE_SERVICE_ROOT
from launchpadlib.credentials import Credentials

SERVICE_ROOT = EDGE_SERVICE_ROOT

# get all uploads (including all distro series and all pockets)
# for this distribution since a given date
DISTRIBUTION = "Ubuntu"


def debug_print_lp(lp_object):
    """ helper to debug the launchpadlib stuff more easily"""
    print("name: %s" % lp_object.__class__.__name__)
    print("attributes: %s" % sorted(lp_object.lp_attributes))
    print("collections: %s" % sorted(lp_object.lp_collections))
    print("entries: %s" % sorted(lp_object.lp_entries))
    print("operations: %s" % sorted(lp_object.lp_operations))
    print("")


def poolhash(name):
    if name.startswith("lib"):
        return name[0:4]
    else:
        return name[0:1]


class LaunchpadSourcePackage:
    """ represents a launchpad source package """

    # arches and pool
    SUPPORTED_ARCHES = ["i386", "amd64"]

    def __init__(self, launchpad, lp_source):
        self._launchpad = launchpad
        self._lp_source = lp_source
        self._srcname = self._lp_source.source_package_name
        self._srcver = self._lp_source.source_package_version
        # srip epoch, just like the other changelog extractors
        if ":" in self._srcver:
            self._srcver = self._srcver.split(":")[1]
        self._srcurls = self._lp_source.sourceFileUrls()
        self._srcomponent = self._lp_source.component_name

    @property
    def published(self):
        " is it already published "
        return self._lp_source.date_published is not None

    @property
    def pending(self):
        return self.status == "Pending"

    @property
    def status(self):
        return self._lp_source.status

    @property
    def srcname(self):
        return self._srcname

    @property
    def srcversion(self):
        return self._srcver

    @property
    def srcurls(self):
        return self._srcurls

    @property
    def srccomponent(self):
        return self._srcomponent

    @property
    def binary_packages_versions_components(self):
        binaries = set()
        for binary in self._lp_source.getPublishedBinaries():
            binaries.add((binary.binary_package_name,
                          binary.binary_package_version,
                          binary.component_name))
        return binaries

    def __str__(self):
        return "%s: %s (%s)" % (self._lp_source.source_package_name,
                                self._lp_source.source_package_version,
                                self._lp_source.date_published)


class LaunchpadChangelogsCrawler:
    """ a crawler that can find out about changed packages in LP """

    # LP service name (free form)
    SERVICE_NAME = 'get_changelogs'

    # local dir layout
    ROOT = os.getcwd()
    CREDENTIALS_FILE = os.path.join(ROOT, "lp-credential.txt")

    # the time we last checked
    LAST_CHECK_FILE = os.path.join(ROOT, "last_check.txt")

    # the files we are interessted in
    CHANGELOG_EXTRACT_FILES = ["copyright", "changelog", "NEWS.Debian"]

    def __init__(self, targetdir="./changelogs"):
        self.lp_cachedir = os.path.abspath(os.path.join(self.ROOT,
                                                        "lp-cache/"))
        self._launchpad = None
        self.targetdir = targetdir
        # statistics
        self.skipped = 0
        self.extracted = 0
        # date
        self._time_of_last_check = 0
        if os.path.exists(self.LAST_CHECK_FILE):
            with open(self.LAST_CHECK_FILE, "r") as fd:
                self._time_of_last_check = float(fd.read().strip())
            # be paraoid and substract 2h, it seems like we miss
            # packages otherwise (LP: #685814, #372270)
            self._time_of_last_check -= 2 * 60 * 60
        else:
            logging.warn("assuming last check 30 days ago")
            logging.warn("MAKE SURE THAT YOU POPULATED THE CHANGELOGS "
                         "FROM A DIFFERENT SOURCE INITIALLY")
            self._time_of_last_check = time.time() - 30 * 24 * 60 * 60

    def login(self):
        " login and figure out if interactive or token login can be used"
        if not os.path.exists(self.CREDENTIALS_FILE):
            self.login_interactive()
        else:
            self.login_with_token(self.CREDENTIALS_FILE)

    def login_interactive(self):
        service = SERVICE_ROOT
        self._launchpad = Launchpad.get_token_and_login(self.SERVICE_NAME,
                                                        service,
                                                        self.lp_cachedir)
        with open(self.CREDENTIALS_FILE, "w+") as fd:
            self._launchpad.credentials.save(fd)

    def login_with_token(self, credentials_file):
        service = SERVICE_ROOT
        credentials = Credentials()
        with open(credentials_file, "r") as fd:
            credentials.load(fd)
        self._launchpad = Launchpad(credentials, service, self.lp_cachedir)

    def _fetch_source(self, srcurls, tmpdir):
        for url in srcurls:
            target = os.path.join(
                tmpdir,
                urllib.unquote(os.path.basename(urllib.splithost(url)[1])))

            with open(target, "w+") as fd:
                try:
                    fd.write(urllib2.urlopen(url).read())
                except urllib2.HTTPError:
                    return False
        return True

    def _unpack_source(self, unpackdir, tmpdir):
        dsc = glob.glob("%s/*.dsc" % tmpdir)
        assert len(dsc) == 1
        # for preexec_fn see http://bugs.python.org/issue1652
        with open(os.path.devnull, "w") as devnull:
            res = subprocess.call(["dpkg-source", "--no-check", "-x", dsc[0],
                                   unpackdir],
                                  stdout=devnull,
                                  preexec_fn=lambda: signal.signal(
                                  signal.SIGPIPE,
                                  signal.SIG_DFL))
        return (res == 0)

    def get_changelogs(self):
        ubuntu = self._launchpad.distributions[DISTRIBUTION]
        archive = ubuntu.main_archive
        date = datetime.datetime.fromtimestamp(self._time_of_last_check)
        logging.info("last checked: %s" % date)
        self.get_changelogs_since(archive, date.isoformat())
        return True

    def get_changelogs_since(self, archive, date):
        self._time_of_last_check = time.time()
        now = time.time()
        changed = archive.getPublishedSources(created_since_date=date)
        logging.debug("getPublishedSources() took %i seconds" %
                      (time.time() - now))
        self._get_changelogs_from_source_package_history_collection(changed)

    def _unpack_changelogs_to_target(self, s, dest):
        # check if we have the data for this already
        if os.path.exists(os.path.join(dest, "changelog")):
            logging.debug("skipping already existing '%s'" % dest)
            self.skipped += 1
            return False

        # fetch/unpack
        tmpdir = tempfile.mkdtemp(prefix="lp-crawler-")
        if not self._fetch_source(s.srcurls, tmpdir):
            logging.error("%s %s failed to fetch" % (s.srcname, s.srcversion))
            return False
        unpackdir = os.path.join(tmpdir, "target/")
        if not self._unpack_source(unpackdir, tmpdir):
            logging.error("%s %s failed to unpack in %s" %
                          (s.srcname, s.srcversion, tmpdir))
            return False

        # extract interessting data
        if os.path.exists(os.path.join(unpackdir, "debian/changelog")):
            logging.info("extracting changelog for %s %s" %
                         (s.srcname, s.srcversion))
            self.extracted += 1
            # add primary dir (based on src data)
            for f in self.CHANGELOG_EXTRACT_FILES:
                src = "%s/debian/%s" % (unpackdir, f)
                if os.path.exists(src):
                    if not os.path.exists(dest):
                        # broken symlink, get rid of them
                        if os.path.islink(dest):
                            os.remove(dest)
                        os.makedirs(dest)
                    logging.debug("extracting '%s' to '%s'" % (src, dest))
                    shutil.copy(src, dest)
        else:
            logging.error("%s - %s has no debian/changelog" %
                          (s.srcname, s.srcversion))
            return False

        # remove unpack dir
        shutil.rmtree(tmpdir)
        return True

    def _get_changelogs_from_source_package_history_collection(self, changed):
        start_time = time.time()
        for source_raw in changed:
            s = LaunchpadSourcePackage(self._launchpad, source_raw)
            logging.debug("source package: '%s'" % s)

            if s.pending:
                logging.info("%s %s in PENDING state" %
                             (s.srcname, s.srcversion))
            if not s.published:
                logging.debug("%s in state %s" % (s.srcname, s.status))
                continue

            # main extraction
            dest = "%s/pool/%s/%s/%s/%s_%s" % (self.targetdir,
                                               s.srccomponent,
                                               poolhash(s.srcname),
                                               s.srcname,
                                               s.srcname,
                                               s.srcversion)

            if not self._unpack_changelogs_to_target(s, dest):
                continue

            # Separate symlink allowing for direct binary + version lookup
            # without requiring the binary component too
            for (binary, version, comp) in \
                    s.binary_packages_versions_components:
                lnkdir = "%s/binary/%s/%s" % (self.targetdir,
                                              poolhash(binary),
                                              binary)
                lnk = "%s/%s" % (lnkdir, version)
                if not os.path.exists(lnk):
                    # broken symlink, get rid of them
                    if os.path.islink(lnk):
                        os.remove(lnk)
                    logging.debug("create compat symlink '%s' -> '%s'" %
                                  (dest, lnk))
                    if not os.path.exists(lnkdir):
                        os.makedirs(lnkdir)
                    os.symlink(os.path.abspath(dest), lnk)

        logging.debug("extracted %i changelogs in %i seconds" % (
            self.extracted, time.time() - start_time))
        return True

    def write_last_check_date(self):
        with open(self.LAST_CHECK_FILE, "w+") as fd:
            fd.write("%s" % self._time_of_last_check)

if __name__ == "__main__":
    if "--debug" in sys.argv:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    lock = apt_pkg.get_lock("lp-extract-changelogs.lock")
    if lock < 0:
        logging.warn("another extractor is running, skipping")
        sys.exit(1)

    c = LaunchpadChangelogsCrawler()
    c.login()
    if c.get_changelogs():
        c.write_last_check_date()
    logging.info("skipped: %i, extracted: %i" % (c.skipped, c.extracted))

    # test code
    #ubuntu = c._launchpad.distributions[DISTRIBUTION]
    #archive = ubuntu.main_archive
    #c.get_changelogs_from_source_packages(archive, ["acct"])
